# -*- coding: utf-8 -*-
"""
This module contains the set of Blue Ox's exceptions

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""


class Error(Exception):
    """This is an ambiguous error that occured."""
    pass
